;;(c) 2020 NedCAD
;; FlatMesh first apeared on vanderworp.org and is, will be transferred to nedcad.nl. Search the net for Flatmesh and domain and you'll find it.
;; Anyhow, this version lives here: https://vanderworp.org/en/flattening-ruled-surfaces-to-blanks/
;; Version 1: Some rough code, intentionally not optimized.

(defun flatmesh ( / *rail-A* *rail-B* *loops* *t-start*)

	(setq *t-start* (getvar "millisecs"))
	(if (not (= (getvar 'vendorname) "Bricsys"))
		(progn
			(defun acos (r / ) (atan (sqrt (- 1 (expt r 2))) r))
			(princ "\nAutoCAD misses a function for arccosine. Function added. ")
		)
	)

	(defun create-rail-lists ( / ent j k m n rails rails-half-len rails-len)
		(while (not j)
			(princ "\nThis command is intended for use with a '3D Polygon Mesh', 1 unit wide, such as can be found in a 'Ruled Surface', which is a polyline with DXF Group 'Polyline flag' 70 set to 16 and group 'M-vertex count' 71 set to 2. It is always good to know a bit more about your CAD-program. Possible commands for making a ruled surface: SurfTab1 and RuleSurf.\nSelect an object: ")
			(if
				(and
					(setq ent (car (entsel)))
					(setq ent (entget ent))
					(= 16 (cdr (assoc 70 ent)))
					(= 2 (cdr (assoc 71 ent)))
					(= "POLYLINE" (cdr (assoc 0 ent)))
				)
				(setq j T)
				(princ "\nNothing selected or not a suitable entity type. Try again... ")
			)
		)
		(setq m T)
		(while m
			(setq ent (entget (entnext (cdr (car ent)))))
			(if (/= (cdr (assoc 0 ent)) "SEQEND")
				(progn
					(setq rails (append rails (list (cdr (assoc 10 ent)))))
				)
				(setq m nil)
			)
		)
		(setq k 0 n 1 rails-len (length rails) rails-half-len (/ rails-len 2))
		(while n
			(cond
				((> n rails-len)
					(setq n nil)
				)
				((<= n rails-half-len)
					(setq
						*rail-A* (append *rail-A* (list (nth k rails)))
						n (1+ n)
						k (1+ k)
					)
				)
				((> n rails-half-len)
					(setq
						*rail-B* (append *rail-B* (list (nth k rails)))
						n (1+ n)
						k (1+ k)
					)
				)
				(T nil)
			)
		)
	)

	(defun dist3d (a b / ) (sqrt (+ (expt (- (car b) (car a)) 2) (expt (- (cadr b) (cadr a)) 2) (expt (- (caddr b) (caddr a)) 2))))

	(defun make-blank-pythagoras ( / Bx By cosAB sinAB aAB sigmaX sigmaY Cx Cy Dx Dy CxRo CyRo DxRo DyRo A Ar C Cr aAC aC aCD aBD aD aDC Allpts Apts B Br D Dr Bpts d-eq-pt dAC dAB dAD dBC dCD dBD n)
		(setq
			d-eq-pt 1e-10
			Ar (nth 0 *rail-A*)
			Br (nth 0 *rail-B*)
			Cr (nth 1 *rail-A*)
			Dr (nth 1 *rail-B*)
			dAB (dist3d Ar Br)
			dAC (dist3d Ar Cr)
			dAD (dist3d Ar Dr)
			dBC (dist3d Br Cr)
			dBD (dist3d Br Dr)
			dCD (dist3d Cr Dr)
			A (list 0.0 0.0)
			baseX (car Ar)
			baseY (cadr Ar)
		)
		(if (> d-eq-pt dAB)
			(progn
				(setq
					B A
					Dx dAD
					Dy 0.0
					D (list Dx Dy)
					Cx (/ (- (+ (expt dAD 2) (expt dAC 2)) (expt dCD 2)) (* 2.0 dAD))
					Cy (sqrt (- (expt dAC 2) (expt Cx 2)))
					C (list Cx Cy)
					Apts (list (list baseX baseY) (list (+ baseX Cx) (+ baseY Cy)))
					Bpts (list (list baseX baseY) (list (+ baseX Dx) baseY))
				)
			)
			(progn
				(setq
					B (list dAB 0.0)
					Cx (/ (- (+ (expt dAB 2) (expt dAC 2)) (expt dBC 2)) (* 2.0 dAB))
					Cy (sqrt (- (expt dAC 2) (expt Cx 2)))
					C (list Cx Cy)
					Dx (/ (- (+ (expt dAB 2) (expt dBD 2)) (expt dAD 2)) (* -2.0 dAB))
					Dy (sqrt (- (expt dBD 2) (expt Dx 2)))
					D (list (+ Dx dAB) Dy)
					Apts (list (list baseX baseY) (list (+ baseX Cx) (+ baseY Cy)))
					Bpts (list (list (+ baseX dAB) baseY) (list (+ baseX Dx dAB) (+ baseY Dy)))
				)
			)
		)
		(setq
			sigmaX Cx
			sigmaY Cy
			aAB (angle C D)
			n 2
		)
		(while (< n (length *rail-A*))
			(setq
				cosAB (cos aAB)
				sinAB (sin aAB)
				Ar Cr
				Br Dr
				Cr (nth n *rail-A*)
				Dr (nth n *rail-B*)
				dAB (dist3d Ar Br)
				dAC (dist3d Ar Cr)
				dAD (dist3d Ar Dr)
				dBC (dist3d Br Cr)
				dBD (dist3d Br Dr)
				dCD (dist3d Cr Dr)
				Bx dAB
				By 0.0
				B (list Bx By)
				Cx (/ (- (+ (expt dAB 2) (expt dAC 2)) (expt dBC 2)) (* 2.0 dAB))
				Cy (sqrt (- (expt dAC 2) (expt Cx 2)))
				C (list Cx Cy)
				Dx (/ (- (+ (expt dAB 2) (expt dBD 2)) (expt dAD 2)) (* -2.0 dAB))
				Dy (sqrt (- (expt dBD 2) (expt Dx 2)))
				Dx (+ dAB Dx)
				CxRo (- (* Cx cosAB) (* Cy sinAB))
				CyRo (+ (* Cy cosAB) (* Cx sinAB))
				DxRo (- (* Dx cosAB) (* Dy sinAB))
				DyRo (+ (* Dy cosAB) (* Dx sinAB))
				aAB (angle (list CxRo CyRo) (list DxRo DyRo))
				sigmaX (+ sigmaX CxRo)
				sigmaY (+ sigmaY CyRo)
				C (list (+ sigmaX baseX) (+ sigmaY baseY))
				D (list (+ sigmaX baseX (- DxRo CxRo)) (+ sigmaY baseY (- DyRo CyRo)))
				Apts (append Apts (list C))
				Bpts (append Bpts (list D))
				n (1+ n)
			)
		)
		(setq
			Allpts (append Apts (reverse Bpts))
			*loops* n
		)

		(entmakex
			(append
				(list
					'(0 . "LWPOLYLINE")
					'(100 . "AcDbEntity")
					'(100 . "AcDbPolyline")
					(cons 90 (length Allpts))
					'(70 . 1)
				)
				(mapcar '(lambda (x) (cons 10 x)) Allpts)
			)
		)
	)

	(defun make-blank-rule-cosine ( / A0 A0r A1 A1r aA0A1 aA1 aA1B1 aB0B1 aB1 aB1A1 Allpts Apts B0 B0r B1 B1r Bpts d-eq-pt dA0A1 dA0B0 dA0B1 dA1B0 dA1B1 dB0B1 n)
		(setq
			d-eq-pt 1e-10
			A0r (nth 0 *rail-A*)
			B0r (nth 0 *rail-B*)
			A1r (nth 1 *rail-A*)
			B1r (nth 1 *rail-B*)
			dA0B0 (dist3d A0r B0r)
			dA0A1 (dist3d A0r A1r)
			dB0B1 (dist3d B0r B1r)
			dA0B1 (dist3d A0r B1r)
			dA1B0 (dist3d B0r A1r)
			dA1B1 (dist3d A1r B1r)
			A0 (list (car A0r) (cadr A0r) 0)
		)
		(if (> d-eq-pt dA0B0)
			(progn
				(setq
					B0 A0
					B1 (list (+ (car B0) dB0B1) (cadr A0) 0)
					aB1A1 (- pi (acos (/ (- (+ (expt dB0B1 2) (expt dA1B1 2)) (expt dA0A1 2)) (* 2 dB0B1 dA1B1))))
					A1 (polar B1 aB1A1 dA1B1)
				)
			)
			(progn
				(setq
					B0 (list (+ (car A0) dA0B0) (cadr A0) 0)
					aA0A1 (acos (/ (- (+ (expt dA0A1 2) (expt dA0B0 2)) (expt dA1B0 2)) (* 2 dA0A1 dA0B0)))
					aB0B1 (- pi (acos (/ (- (+ (expt dB0B1 2) (expt dA0B0 2)) (expt dA0B1 2)) (* 2 dB0B1 dA0B0))))
					A1 (polar A0 aA0A1 dA0A1)
					B1 (polar B0 aB0B1 dB0B1)
				)
			)
		)
		(setq
			Apts (list A0 A1)
			Bpts (list B0 B1)
			n 2
		)
		(while (< n (length *rail-A*))
			(setq
				aA1B1 (angle A1 B1)
				A0r A1r
				B0r B1r
				A1r (nth n *rail-A*)
				B1r (nth n *rail-B*)
				dA0B0 (distance A0r B0r)
				dA0A1 (distance A0r A1r)
				dB0B1 (distance B0r B1r)
				dA0B1 (distance A0r B1r)
				dA1B0 (distance B0r A1r)
				aA1 (acos (/ (- (+ (expt dA0A1 2) (expt dA0B0 2)) (expt dA1B0 2)) (* 2 dA0A1 dA0B0)))
				aB1 (- pi (acos (/ (- (+ (expt dB0B1 2) (expt dA0B0 2)) (expt dA0B1 2)) (* 2 dB0B1 dA0B0))))
				A1 (polar A1 (+ aA1B1 aA1) dA0A1)
				B1 (polar B1 (+ aA1B1 aB1) dB0B1)
				Apts (append Apts (list A1))
				Bpts (append Bpts (list B1))
				n (1+ n)
				*loops* n
			)
		)
		(setq Allpts (append Apts (reverse Bpts)))
		(entmakex
			(append
				(list
					'(0 . "LWPOLYLINE")
					'(100 . "AcDbEntity")
					'(100 . "AcDbPolyline")
					(cons 90 (length Allpts))
					'(70 . 1)
				)
				(mapcar '(lambda (x) (cons 10 x)) Allpts)
			)
		)
	)

	(create-rail-lists)
	; (make-blank-rule-cosine)
	(make-blank-pythagoras)
	(princ
		(strcat
			"\n\nFlatMesh is finished. \n\nCountless operations like divisions and additions are executed. To give an example, this job did: "
			(itoa *loops*) " cosines, "
			(itoa *loops*) " sines, "
			(itoa (+ 8 (* 8 *loops*))) " square roots and "
			(itoa (+ 28 (* 28 *loops*))) " exponents. "
			"\n\nFlatMesh finished this task in "
			(rtos (/ (- (getvar "millisecs") *t-start*) 1000.0) 2 3)
			" seconds. \n\nHappy Drafting and don't forget to send pictures to https://vanderworp.org/en/flattening-ruled-surfaces-to-blanks/ ;-)"
		)
	)
	(princ)
)

